/****************************************************************************
 * 
 * tcpurify - main.c
 * 
 * This program is Copyright 2000, 2001 by Ethan Blanton; please see the
 * accompanying copy of the GNU General Public License for your rights
 * regarding the usage and distribution of this software.
 * 
 ***************************************************************************/

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <pcap.h>
#include <netinet/in.h>
#include <netdb.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include "tcpurify.h"
#include "globals.h"
#include "randtable.h"
#include "dump.h"
#include "capture.h"

void cleanup(int signum);
void usage();

/* This lives in libpcap, but we use it to print out our -v and -h messages
 * to let the user know what we were compiled against. */
extern char pcap_version[];

/* The following three structures are the global variables defined in
 * globals.h.  config contains our configuration information (duh), networks
 * contains the information on requested address transformations, and times
 * is used for realtime playback. */
struct Config config = {
  0, 1, 0, 0, 0, NULL, NULL, NULL, NULL, NULL, NULL
};

struct Network networks[NETWORKS];
struct Times times;

int main(int argc, char *argv[])
{
  int opt, count = -1, i;
  char err[PCAP_ERRBUF_SIZE], tmpfname[32], subnet[16];
  struct in_addr subnet_addr;
  uint32_t netmask, xformmask;
  FILE *fp;
  
  for(i = 0; i < NETWORKS; i++) {
    networks[i].subnet = 0x0;
  }
  
  while((opt = getopt(argc, argv, "i:f:o:m:c:T:dzxtrhv")) != -1) {
    switch(opt) {
     case 'i':
      config.interface = optarg;
      break;
     case 'f':
      config.filename = optarg;
      break;
     case 'o':
      config.outfile = optarg;
      break;
     case 'm':
      config.mapfile = optarg;
      break;
     case 'c':
      count = atoi(optarg);
      if(count < 0) {
	fprintf(stderr, "Invalid packet count %d\n", count);
	return(1);		/* Bye-bye */
      }
      break;
     case 'T':
      config.maxtime = atoi(optarg);
      break;
     case 'd':
      config.debug++;
      break;
     case 'x':
      config.truncate = 0;
      break;
     case 't':
      config.timed = 1;
      break;
     case 'r':
      config.reverse = 1;
      break;
     case 'h':
      usage();
      /* Never reached */
      break;
     case 'v':
      printf("tcpurify %s, libpcap %s\n", VERSION, pcap_version);
      return(0);		/* Exit the program */
     default:
      usage();
    }
  }
  if(config.filename && config.interface) {
    fprintf(stderr, "The -f option cannot be used with the -i option\n");
    return(1);			/* Exit 1 */
  }
  if(config.timed && !config.filename) {
    fprintf(stderr, "tcpurify: -t requires the -f option\n");
    return(1);			/* Doh, unsuccessful termination */
  }
  if(config.mapfile && optind < argc) {
    fprintf(stderr, "tcpurify: -m cannot be used with subnet triplets\n");
    return(1);			/* And another termination */
  }
  if (config.reverse && !config.mapfile) {
    fprintf (stderr, "tcpurify: -r requires -m\n");
    return (1);
  }
  
  /* This particularly ugly chunk of code determines what network addresses
   * we wish to transform when we're munging the headers in capture_cb().
   * The address computation below should be changed to use gethostbyname()
   * at some point in the future. */
  /* Update, 12/4/2000 - I'm using inet_pton() now, this is less ugly.  :-) */
  for(opt = optind, i = 0; opt < argc && i < NETWORKS; opt++) {
    if(sscanf(argv[opt], "%15[0-9.]/%x/%x", subnet, 
	      &netmask, &xformmask) != 3) {
      fprintf(stderr, "Invalid triplet %s\n", argv[opt]);
      continue;
    }
    if (inet_pton (AF_INET, subnet, &subnet_addr) < 0) {
      fprintf (stderr, "Invalid address %s: %s\n",
	       subnet, strerror (errno));
      continue;
    }
    networks[i].netmask = netmask;
    networks[i].xformmask = xformmask;
    memcpy (&networks[i].subnet, &subnet_addr, sizeof(subnet_addr));
    if(!table_create(&networks[i])) {
      networks[i].subnet = 0;
      fprintf(stderr, "Invalid triplet %s\n", argv[opt]);
      continue;
    }
    i++;
  }

  /* If you need to increase this constant, it's in tcpurify.h */
  if(i == NETWORKS) {
    fprintf(stderr, "tcpurify: No more than %d transformation triplets can be specified\n",
	    NETWORKS);
    return(1);			/* Exit...  Maybe we should go on with the
				 * first NETWORKS connections? */
  }
  
  if(i) {
    snprintf(tmpfname, 32, "/tmp/tcpurify-map-%lu", time(NULL));
    if((fp = fopen(tmpfname, "w")) == NULL) {
      fprintf(stderr, "Could not open map file %s: %s\n",
	      tmpfname, strerror(errno));
    } else {
      /* Ensure that the mapping file is readably only by us before
       * continuing */
      if(chmod(tmpfname, S_IRUSR | S_IWUSR)) {
	fprintf(stderr, "Error chmoding %s: %s\n", tmpfname, strerror(errno));
	unlink(tmpfname);	/* I don't care if this succeeds */
	return(1);		/* Consider this fatal and exit */
      }
      if(config.debug) {
	fprintf(stderr, "Mapping file is %s\n", tmpfname);
      }
      fprintf(fp, "%d\n", i - 1);
      for(; i > 0; i--) {
	if(!table_write(fp, &networks[i - 1])) {
	  break;
	}
      }
      if(ferror(fp)) {
	fprintf(stderr, "Error writing to %s.\n", tmpfname);
      }
      fclose(fp);
    }
  }
  
  if(config.mapfile) {
    if((fp = fopen(config.mapfile, "r")) == NULL) {
      fprintf(stderr, "Could not open mapping file %s: %s\n",
	      config.mapfile, strerror(errno));
      return(1);
    }
    
    fscanf(fp, "%d", &i);
    for(; i >= 0; i--) {
      if(!table_read(fp, &networks[i], config.reverse)) {
	break;
      }
    }
    if(ferror(fp)) {
      fprintf(stderr, "Error reading mapping file %s: %s\n",
	      config.mapfile, strerror(errno));
      return(1);
    }
    fclose(fp);
  }

  /* If we don't clean up properly while using gzipped streams, the end of
   * the stream can be unusable. */
  signal(SIGTERM, cleanup);
  signal(SIGINT, cleanup);
  signal(SIGHUP, cleanup);
  
  if(config.filename) {
    if((config.p = pcap_open_offline(config.filename, err)) == NULL) {
      fprintf(stderr, "Error opening tracefile %s: %s\n", 
	      config.filename, err);
      return(1);		/* Exit the program */
    }
  } else {
    if(config.interface == NULL) {
      if((config.interface = pcap_lookupdev(err)) == NULL) {
	fprintf(stderr, "Could not get capture interface: %s\n", err);
	return(1);		/* Exit the program */
      }
    }
    
    /* Open the given interface, get the 1st 64 bytes, do be promiscuous...
     * I have no idea what the 1000 is.  :-( */
    if((config.p = pcap_open_live(config.interface, PKT_LEN, 
			   1, 1000, err)) == NULL) {
      fprintf(stderr, "Error opening %s for capture: %s\n",
	      config.interface, err);
      return(1);		/* Exit the program */
    }
  }
  
  dump_open();
  
  if(config.timed) {
    gettimeofday(&times.wall, NULL);
    times.virt.tv_sec = 0;
    if(pcap_loop(config.p, count, timed_cb, NULL) < 0) {
      fprintf(stderr, "pcap_loop: %s\n", pcap_geterr(config.p));
    }
  } else {
    if(pcap_loop(config.p, count, capture_cb, NULL) < 0) {
      fprintf(stderr, "pcap_loop: %s\n", pcap_geterr(config.p));
    }
  }
  dump_close();
  
  return(0);
}

void cleanup(int signum)
{
  dump_close ();
  
  if(config.p) {
    pcap_close(config.p);
  }
  exit(0);
}

/* I think this should be obvious enough. */
void usage()
{
  fprintf(stderr, "tcpurify version %s\n", VERSION);
  fprintf(stderr, "libpcap version %s\n", pcap_version);
  fprintf(stderr, "\
Usage: tcpurify [OPTION] {SUBNET/NETMASK/XFORMMASK}\n\
  -d\t\tdebug\n\
  -x\t\tdisable IP packet truncation\n\
  -r\t\treverse previous mapping (requires -m)\n\
  -i interface\tread input from device interface\n\
  -f filename\tread input from file filename\n\
  -o filename\tsend output to file filename\n\
  -m filename\tuse mapping file filename\n\
  -c count\tcapture only count packets\n\
  -h\t\tdisplay this message and exit\n\
  -v\t\tdisplay version information and exit\n\
\n\
Up to %d SUBNET/NETMASK/XFORMMASK triplets can be specified.  NETMASK and\n\
XFORMMASK are specified as hexadecimal masks; NETMASK is a typical\n\
ifconfig-type netmask, and XFORMMASK indicates the bits that you would like\n\
to randomize by flagging them with a 1.  For example:\n\
\ttcpurify 172.16.0.0/0xfff00000/0xff\n\
Would match the network 172.16.0.0/255.240.0.0, and transform the last eight\n\
bits of the matching addresses.\n\
", NETWORKS);
  exit(1);
}
