/****************************************************************************
 * 
 * tcpurify - randtable.c
 * 
 * This program is Copyright 2000, 2001 by Ethan Blanton; please see the
 * accompanying copy of the GNU General Public License for your rights
 * regarding the usage and distribution of this software.
 * 
 ***************************************************************************/

#include <netinet/in.h>
#include <stdlib.h>
#include <inttypes.h>
#include <stdio.h>
#include <time.h>
#include "randtable.h"

/*
 * Assuming that mask is a 32-bit mask of the "significant" digits in an
 * integer x (i.e. only the bits you will care about in x are 1's), this
 * squishes x into the rightmost n bits needed to represent x uniquely 
 */
uint32_t squish(uint32_t x, uint32_t mask)
{
  uint32_t i, j, y;
  
  j = y = 0;
  
  for(i = 0; i < 32; i++) {
    if(mask & 0x1) {
      y = y | ((x & mask & 0x1) << j);
      j++;
    }
    x = x >> 1;
    mask = mask >> 1;
  }
  
  return(y);
}

/*
 * See squish and think the opposite.  Obviously.
 */
uint32_t unsquish(uint32_t x, uint32_t mask)
{
  uint32_t i, y;
  
  y = 0;
  
  for(i = 0; i < 32; i++) {
    if(mask & 0x1) {
      y = y | ((x & mask & 0x1) << i);
      x = x >> 1;
    }
    mask = mask >> 1;
  }
  
  return(y);
}

/*
 * Read the description of squish before, and then think about this function.
 * :-)  This takes a table of integers in which you only care about the
 * entries that are unique under the mask 'mask' and reduces it to the
 * minimum possible size.
 */
uint32_t *table_compress(uint32_t *table, uint32_t mask) {
  uint32_t *compressed, i, size_comp, size;
  
  size_comp = squish(mask, mask) + 1;
  size = (0x1 << mask_bits(mask)) + 1;
  
  compressed = (uint32_t *)malloc(size_comp * sizeof(uint32_t));
  
  for(i = 0; i < size_comp; i++) {
    compressed[i] = table[unsquish(i, mask)];
  }
  
  return(compressed);
}

/*
 * Read unsquish, then read table_compress and think backwards.
 */
uint32_t *table_expand(uint32_t *compressed, uint32_t mask) {
  uint32_t *table, i, size_comp, size;
  
  size_comp = squish(mask, mask) + 1;
  size = 0x1 << (mask_bits(mask) + 1);
  
  table = (uint32_t *)malloc(size * sizeof(uint32_t));
  
  for(i = 0; i < size_comp; i++) {
    table[unsquish(i, mask)] = compressed[i];
  }
  
  return(table);
}

/*
 * This takes a properly filled-out Network structure and creates a random
 * table in network->table for address translation.
 */
int table_create(struct Network *network)
{
  uint32_t i, size, *temptable;
  
  if(mask_bits(network->xformmask) > 20 
     || mask_bits(network->xformmask) == 0) {
    return(0);			/* This is just too large; ARBITRARY LIMIT */
  }
  
  network->netmask = htonl(network->netmask);
  
  size = squish(network->xformmask, network->xformmask) + 1;
  size = 2 * (0x1 << mask_bits(network->xformmask));
  
  temptable = (uint32_t *)malloc(size * sizeof(uint32_t));
  for(i = 0; i < size; i++) {
    temptable[i] = htonl(unsquish(i, network->xformmask));
  }
  shuffle(temptable, size);
  network->table = table_expand(temptable, network->xformmask);
  free(temptable);
  
  network->xformmask = htonl(network->xformmask);
  
  return(1);
}

/*
 * table_write compresses a table appropriately and then writes it to stdout
 * with the necessary information from its Network structure to read it back
 * in with no loss of data.
 */
int table_write(FILE *fp, const struct Network *network)
{
  uint32_t mask, size, *table, i;
  
  mask = ntohl(network->xformmask);
  size = squish(mask, mask) + 1;
  
  table = table_compress(network->table, mask);
  fprintf(fp, "\n%x\n", ntohl(network->subnet));
  fprintf(fp, "%x\n", ntohl(network->netmask));
  fprintf(fp, "%x\n", mask);
  for(i = 0; i < size; i++) {
    fprintf(fp, "%x\n", ntohl(table[i]));
  }
  free(table);
  
  if(ferror(fp)) {
    return(0);
  }
  return(1);
}

/*
 * This reads in the table written with table_write, reversing it as it goes
 */
int table_read(FILE *fp, struct Network *network, int reverse)
{
  uint32_t i, size, temp, mask, *table;
  
  fscanf(fp, "%x", &temp);
  network->subnet = htonl(temp);
  fscanf(fp, "%x", &temp);
  network->netmask = htonl(temp);
  fscanf(fp, "%x", &mask);
  network->xformmask = htonl(mask);
  
  size = squish(mask, mask) + 1;
  table = (uint32_t *)malloc(size * sizeof(uint32_t));
  
  for(i = 0; i < size; i++) {
    fscanf(fp, "%x", &temp);
    if (reverse) {
      table[temp] = htonl (i);
    } else {
      table[i] = htonl (temp);
    }
  }
  
  network->table = table_expand(table, mask);
  free(table);

  if(ferror(fp)) {
    return(0);
  }
  return(1);
}

/*
 * Tell me the most significant 1 in mask
 */
int mask_bits(uint32_t mask)
{
  int i, bits = 0;
  
  for(i = 0; i < 32; i++) {
    if(mask & (0x1 << i)) {
      bits = i;
    }
  }
  
  return(bits);
}

/*
 * If you don't recognize this, you shouldn't be reading this source.  ;-)
 */
void swap(uint32_t *table, uint32_t x, uint32_t y) {
  uint32_t temp;
  
  temp = table[x];
  table[x] = table[y];
  table[y] = temp;
}

/*
 * A simple randomization function that shuffles entries 1 .. size - 2
 * of the given table.  It leaves the first and last entries for the reasons
 * discussed in the accompanying README file.
 */
void shuffle(uint32_t *table, uint32_t size) {
  uint32_t i, j, newpos;
  
  srand(time(NULL));
  for(i = 0; i < 3; i++) {
    for(j = 1; j < size - 1; j++) {
      newpos = 1 + (int)((double)(size - 2) * rand() / (RAND_MAX + 1.0));
      swap(table, j, newpos);
    }
  }
}
